#ifndef MPC_H_INCLUDED
#define MPC_H_INCLUDED

#include <map>
#include <vector>
#include <cmath>
#include <algorithm>
#include <Eigen/Core>
#include <Eigen/Geometry>
#include <assert.h>
#include <numeric>
#include <random>
#include <iterator>
#include <ros/ros.h>
#include <ros/node_handle.h>
#include <XmlRpcValue.h>

class MPC
{
public:
    MPC(bool gt, bool w);
    ~MPC();
    std::vector<int> T;
    std::vector<int> control_dim;
    Eigen::MatrixXd A, B, C, D;
    Eigen::MatrixXd Q, R;
    Eigen::MatrixXd Mx, Mc;
    Eigen::MatrixXd K;

protected:
    void init_parameters();
    void build_ABCD();
    void get_K();
    void get_mats();
    void cost_matrices();
    void constraints_matrices();
    //static double cost(const Eigen::VectorXd& c, const Eigen::VectorXd& x, const Eigen::VectorXd& xref, const Eigen::VectorXd& uref);
    void optimize(const Eigen::VectorXd& x, const Eigen::VectorXd& xref, const Eigen::VectorXd& uref);

private:
    bool ground_truth;
    bool wind;
    int N;
    double delta_t;
    double lxy_kp, lxy_ki, lxy_kd;
    double lz_kp, lz_ki, lz_kd;
    double az_kp, az_ki, az_kd;
    double lxy_tau, lz_tau, az_tau;
    ros::NodeHandle _nh;
    struct state {
      Eigen::Vector3d position;
      Eigen::Vector3d euler_angles;
      double inertia_coeffs[6];
      Eigen::Vector3d linear_speeds;
      Eigen::Vector3d angular_speeds;
      Eigen::Vector3d linear_accelerations;
      Eigen::Vector3d angular_accelerations;
    } s;
    struct S
    {
        static double cost(const Eigen::VectorXd& c, const Eigen::VectorXd& x, const Eigen::VectorXd& xref, const Eigen::VectorXd& uref)
    } J;
};

#endif // MPC_H_INCLUDED
