#include <cstdio>
#include <stdio.h>
#include <iostream>
#include <map>
#include <string>

#include "std_msgs/Int16.h"
#include "std_msgs/Float32MultiArray.h"
#include "geometry_msgs/Pose.h"
#include "geometry_msgs/PoseArray.h"
#include "geometry_msgs/Wrench.h"

#include "mpc/MPC.h"
#include <dlib/optimization.h>
#include <dlib/global_optimization.h>


MPC::MPC(bool gt, bool w)
{
    ground_truth = gt; 
    wind = w; 
    init_parameters();
}

MPC::~MPC() {}


void MPC::build_ABCD()
{
    assert(T.size() == 4);
    
    int n(std::accumulate(T.begin(), T.end(), 0));
    //int p(std::accumulate(control_dim.begin(), control_dim.end(), 0));
    Eigen::MatrixXd _A(n, n), _B(n, 8);
    _A.setZero(n, n); _B.setZero(n, 8);
    double kp[4] = {lxy_kp, lxy_kp, lz_kp, az_kp};
    double ki[4] = {lxy_ki, lxy_ki, lz_ki, az_ki};
    double kd[4] = {lxy_kd, lxy_kd, lz_kd, az_kd};
    double r[4] = {delta_t / (delta_t + lxy_tau), delta_t / (delta_t + lxy_tau), delta_t / (delta_t + lz_tau), delta_t / (delta_t + az_tau)};
    int dim_T = 0;
    int dim_C = 0;
    for (int i(0); i<4; i++)
    {
        Eigen::MatrixXd a(T[i], T[i]), b(1, 2);
        double gamma(kp[i] + ki[i]*delta_t + kd[i]/delta_t);
        int j(std::accumulate(T.begin(), T.begin()+i, 0));
        if (T[i] == 5)
        {
            a << 1, 0, delta_t, 0, delta_t*delta_t/2,
                 0, 1, 0, delta_t, 0,
                 0, 0, 1, 0, delta_t,
                 0, 0, 0, 0, 1,
                 0, kd[i]/delta_t, -gamma, kd[i], -kd[i];
        }
        else
        {
            a << 1, delta_t, delta_t*delta_t/2,
                 0, 1, delta_t,
                 0, -gamma, 0;
        }
        b << (1-r[i]) * gamma - kd[i]/delta_t, r[i] * gamma;
        for (int k(0); k<T[i]; k++)
            for (int m(0); m<T[i]; m++) {_A(dim_T+k, dim_T+m) = a(k, m);}
        for (int t(0); t<2; t++) {_B(dim_T + T[i]-1, dim_C+t) = b(0, t);}
        dim_T += T[i];
        dim_C += 2;
    }
    A = _A; B= _B;
    if (!ground_truth) {Eigen::MatrixXd c(n, n);};
    if (wind) {};
    // ...
}


void MPC::init_parameters() 
{
    _nh.getParam("controller/twist/linear/xy/k_p", lxy_kp);
    _nh.getParam("controller/twist/linear/xy/k_i", lxy_ki);
    _nh.getParam("controller/twist/linear/xy/k_d", lxy_kd);
    _nh.getParam("controller/twist/linear/xy/time_constant", lxy_tau);
    
    _nh.getParam("controller/twist/linear/z/k_p", lz_kp);
    _nh.getParam("controller/twist/linear/z/k_i", lz_ki);
    _nh.getParam("controller/twist/linear/z/k_d", lz_kd);
    _nh.getParam("controller/twist/linear/z/time_constant", lz_tau);
    
    _nh.getParam("controller/twist/angular/z/k_p", az_kp);
    _nh.getParam("controller/twist/angular/z/k_i", az_ki);
    _nh.getParam("controller/twist/angular/z/k_d", az_kd);
    _nh.getParam("controller/twist/angular/z/time_constant", az_tau);
    
    if (lxy_kd == 0.0) {T.push_back(3);} else {T.push_back(5);};
    if (lxy_kd == 0.0) {T.push_back(3);} else {T.push_back(5);};
    if (lz_kd == 0.0) {T.push_back(3);} else {T.push_back(5);};
    if (az_kd == 0.0) {T.push_back(3);} else {T.push_back(5);};
    
    //if (lxy_kd == 0.0) {T.push_back(3); control_dim.push_back(1);} else {T.push_back(5); control_dim.push_back(2);};
    //if (lxy_kd == 0.0) {T.push_back(3); control_dim.push_back(1);} else {T.push_back(5); control_dim.push_back(2);};
    //if (lz_kd == 0.0) {T.push_back(3); control_dim.push_back(1);} else {T.push_back(5); control_dim.push_back(2);};
    //if (az_kd == 0.0) {T.push_back(3); control_dim.push_back(1);} else {T.push_back(5); control_dim.push_back(2);};
    
    _nh.getParam("N", N);
    _nh.getParam("delta_t", delta_t);
    
    build_ABCD();
    get_K();
}


void MPC::get_K()
{
    XmlRpc::XmlRpcValue Kvec;
    _nh.getParam("K", Kvec);
    assert(Kvec.getType() == XmlRpc::XmlRpcValue::TypeArray);
    Eigen::MatrixXd k(12, 8);
    for (int i(0); i<12; i++)
    {
        for (int j(0); j<8; j++)
            k(i, j) = Kvec[i][j];
    }
    K = k;
}


void MPC::get_mats()
{
    XmlRpc::XmlRpcValue matX, matC;
    _nh.getParam("Mx", matX);
    _nh.getParam("Mc", matC);
    assert(matX.getType() == XmlRpc::XmlRpcValue::TypeArray);
    assert(matC.getType() == XmlRpc::XmlRpcValue::TypeArray);
    Eigen::MatrixXd mx(N*12, 12);
    Eigen::MatrixXd mc(N*12, N*8);
    for (int i(0); i<N*12; i++)
    {
        for (int j(0); j<12; j++)
            mx(i, j) = matX[i][j];
        for (int j(0); j<N*8; j++)
            mc(i, j) = matC[i][j];
    }
    Mx = mx;
    Mc = mc;
}


void MPC::cost_matrices()
{
    XmlRpc::XmlRpcValue Qvec, Rvec;
    _nh.getParam("Q", Qvec);
    _nh.getParam("R", Rvec);
    assert(Qvec.getType() == XmlRpc::XmlRpcValue::TypeArray);
    assert(Rvec.getType() == XmlRpc::XmlRpcValue::TypeArray);
    Eigen::MatrixXd q(N*12, N*12), r(N*8, N*8);
    for (int i(0); i<N*12; i++)
    {
        for (int j(0); j<N*12; j++)
            q(i, j) = Qvec[i][j];
    }
    for (int i(0); i<N*8; i++)
    {
        for (int j(0); j<N*8; j++)
            r(i, j) = Rvec[i][j];
    }
    Q = q;
    R = r;
}


void MPC::constraints_matrices()
{
    
}


void MPC::optimize(const Eigen::VectorXd& x, const Eigen::VectorXd& xref, const Eigen::VectorXd& uref)
{
    double result = dlib::find_min_global(J.cost, 
                                          {-10,-10}, // lower bounds
                                          {10,10}, // upper bounds
                                          std::chrono::milliseconds(500) // run this long
                                          );
}
